@@ -14,17 +14,17 @@ use strided::Stride;
 
 use tc_btree::Node;
 use tc_error::*;
-use tc_transact::fs::{BlockData, BlockId, CopyFrom, Dir, File, Persist, Restore};
+use tc_transact::fs::{BlockId, CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{Transact, Transaction, TxnId};
-use tc_value::{Number, NumberClass, NumberInstance, NumberType, Value};
-use tcgeneric::{TCBoxTryFuture, TCTryStream, Tuple};
+use tc_value::{Number, NumberClass, NumberInstance, NumberType};
+use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use crate::stream::{Read, ReadValueAt};
 use crate::transform;
 use crate::{coord_bounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
 
 use super::access::BlockListTranspose;
-use super::{DenseAccess, DenseAccessor, PER_BLOCK};
+use super::{DenseAccess, DenseAccessor, DenseWrite, PER_BLOCK};
 
 const MEBIBYTE: usize = 1_048_576;
 
@@ -115,12 +115,6 @@ where
             .try_fold(0u64, |block_len, size| future::ready(Ok(size + block_len)))
             .await?;
 
-        if file.is_empty(&txn_id).await? {
-            return Err(TCError::unsupported(
-                "tried to create a dense tensor from an empty block list",
-            ));
-        }
-
         let shape = if let Some(shape) = shape {
             if shape.size() != size {
                 return Err(TCError::unsupported(format!(
@@ -140,7 +134,7 @@ where
     }
 
     /// Construct a new `BlockListFile` from the given `Stream` of elements.
-    pub async fn from_values<S: Stream<Item = Number> + Send + Unpin>(
+    pub async fn from_values<S: Stream<Item = TCResult<Number>> + Send + Unpin>(
         file: FD,
         txn_id: TxnId,
         shape: Shape,
@@ -151,8 +145,9 @@ where
 
         let mut i = 0u64;
         let mut size = 0u64;
-        let mut values = values.chunks((Array::max_size() as usize) / dtype.size());
+        let mut values = values.chunks(PER_BLOCK);
         while let Some(chunk) = values.next().await {
+            let chunk = chunk.into_iter().collect::<TCResult<Vec<Number>>>()?;
             size += chunk.len() as u64;
             let block_id = BlockId::from(i);
             let block = Array::from(chunk).cast_into(dtype);
@@ -190,7 +185,8 @@ where
 
         let values = stream::iter(0..shape.size())
             .map(Number::from)
-            .map(|i| start + (i * step));
+            .map(|i| start + (i * step))
+            .map(Ok);
 
         Self::from_values(file, txn_id, shape, dtype, values).await
     }
@@ -198,6 +194,7 @@ where
     /// Consume this `BlockListFile` handle and return a `Stream` of `Array` blocks.
     pub fn into_stream(self, txn_id: TxnId) -> impl Stream<Item = TCResult<Array>> + Unpin {
         let num_blocks = div_ceil(self.size(), PER_BLOCK as u64);
+
         let blocks = stream::iter((0..num_blocks).into_iter().map(BlockId::from))
             .map(move |block_id| self.file.clone().read_block_owned(txn_id, block_id))
             .buffered(num_cpus::get())
@@ -209,7 +206,9 @@ where
     /// Sort the elements in this `BlockListFile`.
     pub async fn merge_sort(&self, txn_id: TxnId) -> TCResult<()> {
         let num_blocks = div_ceil(self.size(), PER_BLOCK as u64);
-        if num_blocks <= 1 {
+        if num_blocks == 0 {
+            return Ok(());
+        } else if num_blocks == 1 {
             let block_id = BlockId::from(0u64);
             let mut block = self.file.write_block(txn_id, block_id).await?;
             block.sort(true)?;
@@ -236,12 +235,7 @@ where
     }
 
     async fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        if !self.shape().contains_coord(&coord) {
-            return Err(TCError::bad_request(
-                "invalid coordinate",
-                Tuple::from(coord),
-            ));
-        }
+        self.shape().validate_coord(&coord)?;
 
         let value = value.into_type(self.dtype());
 
@@ -260,6 +254,31 @@ where
             .set_value(offset as usize, value)
             .map_err(TCError::from)
     }
+
+    async fn overwrite<B: DenseAccess<FD, FS, D, T>>(&self, txn: T, value: B) -> TCResult<()> {
+        if value.shape() != self.shape() {
+            return Err(TCError::unsupported(format!(
+                "cannot overwrite a Tensor of shape {} with one of shape {}",
+                self.shape(),
+                value.shape()
+            )));
+        }
+
+        let txn_id = *txn.id();
+        let num_blocks = div_ceil(value.size(), PER_BLOCK as u64);
+        let contents = value.block_stream(txn).await?;
+        stream::iter((0..num_blocks).map(BlockId::from))
+            .zip(contents)
+            .map(|(block_id, r)| r.map(|array| (block_id, array)))
+            .map_ok(|(block_id, array)| async {
+                let mut block = self.file.write_block(txn_id, block_id).await?;
+                *block = array;
+                Ok(())
+            })
+            .try_buffer_unordered(num_cpus::get())
+            .try_fold((), |_, _| future::ready(Ok(())))
+            .await
+    }
 }
 
 impl<FD: Send, FS: Send, D: Send, T: Send> TensorAccess for BlockListFile<FD, FS, D, T> {
@@ -296,9 +315,14 @@ where
         DenseAccessor::File(self)
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
             let size = self.size();
+            if size == 0 {
+                let blocks: TCBoxTryStream<'a, Array> = Box::pin(stream::empty());
+                return Ok(blocks);
+            }
+
             let file = self.file;
             let block_stream = Box::pin(
                 stream::iter(0..(div_ceil(size, PER_BLOCK as u64)))
@@ -307,7 +331,7 @@ where
                     .map_ok(|block| (*block).clone()),
             );
 
-            let block_stream: TCTryStream<Array> = Box::pin(block_stream);
+            let block_stream: TCBoxTryStream<Array> = Box::pin(block_stream);
             Ok(block_stream)
         })
     }
@@ -348,38 +372,98 @@ where
             })
             .await
     }
+}
 
-    async fn write<B: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: B) -> TCResult<()> {
-        if value.shape() != self.shape() {
+#[async_trait]
+impl<FD, FS, D, T> DenseWrite<FD, FS, D, T> for BlockListFile<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
+    async fn write<B: DenseAccess<FD, FS, D, T>>(
+        &self,
+        txn: Self::Txn,
+        bounds: Bounds,
+        value: B,
+    ) -> TCResult<()> {
+        self.shape().validate_bounds(&bounds)?;
+
+        if bounds == Bounds::all(self.shape()) {
+            return self.overwrite(txn, value).await;
+        }
+
+        let slice_shape = bounds.to_shape(self.shape())?;
+        if &slice_shape != value.shape() {
             return Err(TCError::unsupported(format!(
                 "cannot overwrite a Tensor of shape {} with one of shape {}",
-                self.shape(),
+                slice_shape,
                 value.shape()
             )));
         }
 
+        let rebase = transform::Slice::new(self.shape().clone(), bounds)?;
+        let size = rebase.size();
         let txn_id = *txn.id();
-        let num_blocks = div_ceil(value.size(), PER_BLOCK as u64);
-        let contents = value.block_stream(txn).await?;
-        stream::iter((0..num_blocks).map(BlockId::from))
-            .zip(contents)
-            .map(|(block_id, r)| r.map(|array| (block_id, array)))
-            .map_ok(|(block_id, array)| async {
-                let mut block = self.file.write_block(txn_id, block_id).await?;
-                *block = array;
-                Ok(())
+        let offsets = (0..size)
+            .step_by(PER_BLOCK)
+            .map(|start| {
+                let end = start + PER_BLOCK as u64;
+                if end > size {
+                    Offsets::range(start, size)
+                } else {
+                    Offsets::range(start, end)
+                }
+            })
+            .map(|offsets| Coords::from_offsets(offsets, rebase.shape()))
+            .map(|coords| rebase.invert_coords(&coords).to_offsets(self.shape()));
+
+        let blocks = value.block_stream(txn).await?;
+
+        let af_per_block = af::constant(PER_BLOCK as u64, af::Dim4::new(&[1, 1, 1, 1]));
+        stream::iter(offsets)
+            .zip(blocks)
+            .map(|(offsets, r)| r.map(|array| (offsets, array)))
+            .map_ok(|(offsets, array)| {
+                let af_per_block = af_per_block.clone();
+
+                async move {
+                    let indices: ArrayExt<u64> =
+                        af::modulo(offsets.af(), &af_per_block, true).into();
+
+                    let block_offsets = af::div(offsets.af(), &af_per_block, true);
+                    let block_ids = ArrayExt::from(af::set_unique(&block_offsets, true)).to_vec();
+
+                    let mut start = 0;
+                    for block_id in block_ids.into_iter() {
+                        let af_block_id = ArrayExt::from(&[block_id][..]);
+                        let (len, _) = af::sum_all(&af::eq(&block_offsets, af_block_id.af(), true));
+                        let end = start + len as usize;
+                        let indices = indices.slice(start, end);
+                        let array = array.slice(start, end).map_err(TCError::from)?;
+
+                        let mut block = self.file.write_block(txn_id, block_id.into()).await?;
+
+                        block.set(&indices, &array)?;
+
+                        start = end;
+                    }
+
+                    Ok(())
+                }
             })
             .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
+            .try_fold((), |(), ()| future::ready(Ok(())))
             .await
     }
 
     async fn write_value(&self, txn_id: TxnId, mut bounds: Bounds, value: Number) -> TCResult<()> {
         debug!("BlockListFile::write_value {} at {}", value, bounds);
+        self.shape().validate_bounds(&bounds)?;
 
-        if !self.shape().contains_bounds(&bounds) {
-            return Err(TCError::bad_request("bounds out of bounds", bounds));
-        } else if bounds.len() == self.ndim() {
+        if bounds.len() == self.ndim() {
             if let Some(coord) = bounds.as_coord() {
                 return self.write_value_at(txn_id, coord, value).await;
             }
@@ -432,12 +516,7 @@ where
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
         Box::pin(async move {
-            if !self.shape().contains_coord(&coord) {
-                return Err(TCError::bad_request(
-                    "coordinate is out of bounds",
-                    Value::from_iter(coord),
-                ));
-            }
+            self.shape().validate_coord(&coord)?;
 
             let offset: u64 = coord_bounds(self.shape())
                 .iter()
@@ -480,11 +559,11 @@ where
     T: Transaction<D>,
     B: DenseAccess<FD, FS, D, T>,
 {
-    async fn copy_from(other: B, file: FD, txn: T) -> TCResult<Self> {
+    async fn copy_from(other: B, file: FD, txn: &T) -> TCResult<Self> {
         let txn_id = *txn.id();
         let dtype = other.dtype();
         let shape = other.shape().clone();
-        let blocks = other.block_stream(txn).await?;
+        let blocks = other.block_stream(txn.clone()).await?;
         Self::from_blocks(file, txn_id, Some(shape), dtype, blocks).await
     }
 }
@@ -843,6 +922,7 @@ where
     T: Transaction<D>,
 {
     fn new(source: BlockListFile<FD, FS, D, T>, bounds: Bounds) -> TCResult<Self> {
+        debug!("BlockListFileSlice::new {} {}", source.shape(), bounds);
         let rebase = transform::Slice::new(source.shape().clone(), bounds)?;
         Ok(Self { source, rebase })
     }
@@ -888,7 +968,12 @@ where
         DenseAccessor::Slice(self)
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
+        if self.size() == 0 {
+            let blocks: TCBoxTryStream<'a, Array> = Box::pin(stream::empty());
+            return Box::pin(future::ready(Ok(blocks)));
+        }
+
         let txn_id = *txn.id();
         let file = self.source.file;
         let shape = self.source.schema.shape;
@@ -917,13 +1002,14 @@ where
             })
         });
 
-        let blocks: TCTryStream<Array> = Box::pin(values);
+        let blocks: TCBoxTryStream<Array> = Box::pin(values);
         Box::pin(future::ready(Ok(blocks)))
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let bounds = self.rebase.invert_bounds(bounds);
-        self.source.slice(bounds)
+        let slice = self.source.slice(bounds)?;
+        Ok(slice)
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
@@ -934,84 +1020,6 @@ where
         let coords = self.rebase.invert_coords(&coords);
         self.source.read_values(txn, coords).await
     }
-
-    async fn write<B: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: B) -> TCResult<()> {
-        if self.shape() != value.shape() {
-            return Err(TCError::unsupported(format!(
-                "cannot overwrite a Tensor of shape {} with one of shape {}",
-                self.shape(),
-                value.shape()
-            )));
-        }
-
-        let size = self.size();
-        let rebase = &self.rebase;
-        let txn_id = *txn.id();
-        let offsets = (0..size)
-            .step_by(PER_BLOCK)
-            .map(|start| {
-                let end = start + PER_BLOCK as u64;
-                if end > size {
-                    Offsets::range(start, size)
-                } else {
-                    Offsets::range(start, end)
-                }
-            })
-            .map(|offsets| Coords::from_offsets(offsets, self.shape()))
-            .map(|coords| {
-                rebase
-                    .invert_coords(&coords)
-                    .to_offsets(self.source.shape())
-            });
-
-        let blocks = value.block_stream(txn).await?;
-
-        let af_per_block = af::constant(PER_BLOCK as u64, af::Dim4::new(&[1, 1, 1, 1]));
-        stream::iter(offsets)
-            .zip(blocks)
-            .map(|(offsets, r)| r.map(|array| (offsets, array)))
-            .map_ok(|(offsets, array)| {
-                let af_per_block = af_per_block.clone();
-
-                async move {
-                    let indices: ArrayExt<u64> =
-                        af::modulo(offsets.af(), &af_per_block, true).into();
-                    let block_offsets = af::div(offsets.af(), &af_per_block, true);
-                    let block_ids = ArrayExt::from(af::set_unique(&block_offsets, true)).to_vec();
-
-                    let mut start = 0;
-                    for block_id in block_ids.into_iter() {
-                        let af_block_id = ArrayExt::from(&[block_id][..]);
-                        let (len, _) = af::sum_all(&af::eq(&block_offsets, af_block_id.af(), true));
-                        let end = start + len as usize;
-                        let indices = indices.slice(start, end);
-                        let array = array.slice(start, end).map_err(TCError::from)?;
-
-                        let mut block = self
-                            .source
-                            .file
-                            .write_block(txn_id, block_id.into())
-                            .await?;
-
-                        block.set(&indices, &array)?;
-
-                        start = end;
-                    }
-
-                    Ok(())
-                }
-            })
-            .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |(), ()| future::ready(Ok(())))
-            .await
-    }
-
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        self.shape().validate_bounds(&bounds)?;
-
-        let bounds = self.rebase.invert_bounds(bounds);
-        self.source.write_value(txn_id, bounds, number).await
-    }
 }
 
 impl<FD, FS, D, T> ReadValueAt<D> for BlockListFileSlice<FD, FS, D, T>
