@@ -6,16 +6,18 @@ use afarray::{Array, CoordBlocks, Coords};
 use async_trait::async_trait;
 use destream::de;
 use futures::future::{self, TryFutureExt};
-use futures::stream::{self, StreamExt, TryStreamExt};
+use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use log::debug;
 
 use tc_btree::{BTreeType, Node};
 use tc_error::*;
-use tc_table::{Column, ColumnBound, Merged, TableIndex, TableInstance, TableSchema};
+use tc_table::{
+    Column, ColumnBound, Merged, TableIndex, TableSchema, TableSlice, TableStream, TableWrite,
+};
 use tc_transact::fs::{CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{Transact, Transaction, TxnId};
 use tc_value::{Bound, Number, NumberClass, NumberInstance, NumberType, UInt, Value, ValueType};
-use tcgeneric::{label, Id, Label, TCTryStream};
+use tcgeneric::{label, Id, Label, TCBoxTryStream, Tuple};
 
 use crate::dense::PER_BLOCK;
 use crate::stream::{sorted_coords, Read, ReadValueAt};
@@ -23,7 +25,7 @@ use crate::transform;
 use crate::{AxisBounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
 
 use super::access::SparseTranspose;
-use super::{SparseAccess, SparseAccessor, SparseStream, SparseTensor};
+use super::{SparseAccess, SparseAccessor, SparseStream, SparseTensor, SparseWrite};
 
 const VALUE: Label = label("value");
 const ERR_CORRUPT: &str = "SparseTensor corrupted! Please file a bug report.";
@@ -97,7 +99,7 @@ where
     T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
-    type Slice = SparseTableSlice<FD, FS, D, T>;
+    type Slice = SparseAccessor<FD, FS, D, T>;
     type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
@@ -111,14 +113,27 @@ where
         Ok(filled)
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
-        if self.is_empty(&txn).await? {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        self.shape().validate_axes(&axes)?;
+
+        if axes.is_empty() || self.is_empty(&txn).await? {
             return Ok(Box::pin(stream::empty()));
         }
 
+        let sort = axes.iter().zip(0..axes.len()).any(|(x, y)| x != &y);
+
         let shape = self.shape();
-        let shape = axes.iter().map(|x| shape[*x]).collect();
-        filled_at::<FD, FS, D, T, _>(&txn, shape, axes, self.table).await
+        let shape = axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>();
+        let coords = filled_at::<FD, FS, D, T, _>(&txn, axes, self.table).await?;
+
+        let coords = CoordBlocks::new(coords, shape.len(), PER_BLOCK);
+
+        if sort {
+            let coords = sorted_coords::<FD, FS, D, T, _>(&txn, shape.into(), coords).await?;
+            Ok(Box::pin(coords))
+        } else {
+            Ok(Box::pin(coords))
+        }
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
@@ -130,16 +145,34 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
-        let table = slice_table(self.table.clone(), &bounds)?;
-        let rebase = transform::Slice::new(self.shape().clone(), bounds)?;
-        Ok(SparseTableSlice::new(self, table, rebase))
+        debug!("SparseTable {} slice {}", self.shape(), bounds);
+        self.shape().validate_bounds(&bounds)?;
+        let table_bounds = table_bounds(self.shape(), &bounds)?;
+        if table_bounds.is_empty() {
+            Ok(self.accessor())
+        } else {
+            let rebase = transform::Slice::new(self.shape().clone(), bounds)?;
+            let table = self.table.clone().slice(table_bounds)?;
+            Ok(SparseTableSlice::new(self, table, rebase).accessor())
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         SparseTranspose::new(self, permutation)
     }
+}
 
+#[async_trait]
+impl<FD, FS, D, T> SparseWrite<FD, FS, D, T> for SparseTable<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    Self: SparseAccess<FD, FS, D, T>,
+{
     async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
+        self.shape().validate_coord(&coord)?;
         upsert_value(&self.table, txn_id, coord, value).await
     }
 }
@@ -156,13 +189,7 @@ where
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
         Box::pin(async move {
-            if !self.shape().contains_coord(&coord) {
-                return Err(TCError::bad_request(
-                    "coordinate out of bounds",
-                    Bounds::from(coord),
-                ));
-            }
-
+            self.shape().validate_coord(&coord)?;
             let dtype = self.dtype();
             read_value_at(self.table, txn, coord, dtype).await
         })
@@ -200,15 +227,15 @@ where
     async fn copy_from(
         instance: SparseTensor<FD, FS, D, T, A>,
         store: D,
-        txn: T,
+        txn: &T,
     ) -> TCResult<Self> {
+        let txn_id = *txn.id();
         let shape = instance.shape().clone();
         let dtype = instance.dtype();
         let schema = Schema { shape, dtype };
-        let accessor = SparseTable::create(&store, schema, *txn.id()).await?;
+        let accessor = SparseTable::create(&store, schema, txn_id).await?;
 
-        let txn_id = *txn.id();
-        let filled = instance.accessor.filled(txn).await?;
+        let filled = instance.accessor.filled(txn.clone()).await?;
 
         filled
             .map_ok(|(coord, value)| accessor.write_value(txn_id, coord, value))
@@ -333,7 +360,7 @@ where
     T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
-    type Slice = Self;
+    type Slice = SparseAccessor<FD, FS, D, T>;
     type Transpose = SparseTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
@@ -344,21 +371,32 @@ where
         let rebase = self.rebase;
         let rows = self.table.rows(*txn.id()).await?;
         let filled = rows
-            .and_then(|row| future::ready(expect_row(row)))
+            .map(|r| r.and_then(|row| expect_row(row)))
             .map_ok(move |(coord, value)| (rebase.map_coord(coord), value));
 
         let filled: SparseStream<'a> = Box::pin(filled);
         Ok(filled)
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
-        if self.is_empty(&txn).await? {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseTableSlice::filled_at");
+        self.shape().validate_axes(&axes)?;
+
+        if axes.is_empty() || self.is_empty(&txn).await? {
             return Ok(Box::pin(stream::empty()));
         }
 
         let shape = self.shape();
-        let shape = axes.iter().map(|x| shape[*x]).collect();
-        filled_at::<FD, FS, D, T, _>(&txn, shape, axes, self.table).await
+        let shape = axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>();
+        let source_axes = (0..self.source.ndim()).collect();
+        let rebase = self.rebase;
+        let coords = filled_at::<FD, FS, D, T, _>(&txn, source_axes, self.table).await?;
+        let coords = CoordBlocks::new(coords, self.source.ndim(), PER_BLOCK)
+            .map_ok(move |coords| rebase.map_coords(coords))
+            .map_ok(move |coords| coords.get(&axes));
+
+        let filled_at = sorted_coords::<FD, FS, D, T, _>(&txn, shape.into(), coords).await?;
+        Ok(Box::pin(filled_at))
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
@@ -367,7 +405,7 @@ where
 
     async fn is_empty(&self, txn: &T) -> TCResult<bool> {
         let mut rows = self.table.clone().rows(*txn.id()).await?;
-        rows.try_next().map_ok(|v| v.is_some()).await
+        rows.try_next().map_ok(|v| v.is_none()).await
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
@@ -379,12 +417,6 @@ where
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         SparseTranspose::new(self, permutation)
     }
-
-    async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        self.shape().validate_coord(&coord)?;
-        let source_coord = self.rebase.invert_coord(&coord);
-        self.source.write_value(txn_id, source_coord, value).await
-    }
 }
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseTableSlice<FD, FS, D, T>
@@ -399,13 +431,7 @@ where
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
         Box::pin(async move {
-            if !self.shape().contains_coord(&coord) {
-                return Err(TCError::bad_request(
-                    "coordinate out of bounds",
-                    Bounds::from(coord),
-                ));
-            }
-
+            self.shape().validate_coord(&coord)?;
             let dtype = self.dtype();
             let source_coord = self.rebase.invert_coord(&coord);
             read_value_at(self.table, txn, source_coord, dtype).await
@@ -454,67 +480,67 @@ where
 
 async fn filled_at<'a, FD, FS, D, Txn, T>(
     txn: &Txn,
-    shape: Shape,
     axes: Vec<usize>,
     table: T,
-) -> TCResult<TCTryStream<'a, Coords>>
+) -> TCResult<impl Stream<Item = TCResult<Coord>> + Send + Unpin>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
     FS: File<Node>,
     D: Dir,
     Txn: Transaction<D>,
-    T: TableInstance<FS, D, Txn>,
+    T: TableStream,
+    T::Selection: TableStream,
     D::FileClass: From<TensorType>,
 {
-    debug!("SparseTable::filled_at {:?}", axes);
-
-    let sort = axes.iter().zip(0..axes.len()).any(|(x, y)| x != &y);
-
-    let ndim = axes.len();
+    assert!(!axes.is_empty());
     let coords = table.select(axes.into_iter().map(Id::from).collect())?;
     let coords = coords.rows(*txn.id()).await?;
-    let coords = coords.map(|r| r.and_then(expect_coord));
-    let coords = CoordBlocks::new(coords, ndim, PER_BLOCK);
-
-    if sort {
-        let coords = sorted_coords::<FD, FS, D, Txn, _>(txn, shape, coords).await?;
-        Ok(Box::pin(coords))
-    } else {
-        Ok(Box::pin(coords))
-    }
+    Ok(coords.map(|r| r.and_then(expect_coord)))
 }
 
-fn slice_table<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>(
-    table: T,
-    bounds: &Bounds,
-) -> TCResult<T::Slice> {
+fn table_bounds(shape: &Shape, bounds: &Bounds) -> TCResult<tc_table::Bounds> {
+    assert!(bounds.len() <= shape.len());
     use AxisBounds::*;
 
     let mut table_bounds = HashMap::new();
     for (axis, axis_bound) in bounds.to_vec().into_iter().enumerate() {
-        let axis: Id = axis.into();
         let column_bound = match axis_bound {
-            At(x) => ColumnBound::Is(u64_into_value(x)),
+            At(x) => Some(ColumnBound::Is(u64_into_value(x))),
+            In(range) if range == (0..shape[axis]) => None,
             In(range) => {
                 let start = Bound::In(u64_into_value(range.start));
                 let end = Bound::Ex(u64_into_value(range.end));
-                (start, end).into()
+                Some((start, end).into())
+            }
+            Of(indices) => {
+                return Err(TCError::bad_request(
+                    "cannot select non-sequential indices from a sparse Tensor",
+                    Tuple::from(indices),
+                ))
             }
-            _ => todo!(),
         };
 
-        table_bounds.insert(axis, column_bound);
+        if let Some(column_bound) = column_bound {
+            table_bounds.insert(axis.into(), column_bound);
+        }
     }
 
-    table.slice(table_bounds.into())
+    Ok(table_bounds.into())
 }
 
-async fn read_value_at<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>(
+async fn read_value_at<D, Txn, T>(
     table: T,
     txn: Txn,
     coord: Coord,
     dtype: NumberType,
-) -> TCResult<(Coord, Number)> {
+) -> TCResult<(Coord, Number)>
+where
+    D: Dir,
+    Txn: Transaction<D>,
+    T: TableSlice,
+    T::Slice: TableStream,
+    <T::Slice as TableStream>::Selection: TableStream,
+{
     let selector: HashMap<Id, ColumnBound> = coord
         .iter()
         .enumerate()
@@ -532,19 +558,20 @@ async fn read_value_at<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInsta
     Ok((coord, value))
 }
 
-async fn upsert_value<F: File<Node>, D: Dir, Txn: Transaction<D>, T: TableInstance<F, D, Txn>>(
-    table: &T,
-    txn_id: TxnId,
-    coord: Coord,
-    value: Number,
-) -> TCResult<()> {
-    let coord = coord.into_iter().map(Number::from).map(Value::Number);
+async fn upsert_value<T>(table: &T, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()>
+where
+    T: TableWrite,
+{
+    let coord = coord
+        .into_iter()
+        .map(Number::from)
+        .map(Value::Number)
+        .collect();
 
     if value == value.class().zero() {
-        let key = (0..coord.len()).map(Id::from).zip(coord).collect();
-        table.delete_row(txn_id, key).await
+        table.delete(txn_id, coord).await
     } else {
-        let key = coord.collect();
+        let key = coord;
         table.upsert(txn_id, key, vec![Value::Number(value)]).await
     }
 }
