@@ -35,6 +35,8 @@ where
         .await?;
 
     let offsets = sort_coords::<FD, FS, D, T, _>(file, txn_id, coords, shape.clone()).await?;
+    debug!("sorted coords");
+
     let offsets = offsets
         .into_stream(txn_id)
         .map_ok(|array| array.type_cast());
@@ -62,6 +64,7 @@ where
         "sort values by coordinate for Tensor with shape {}",
         source.shape()
     );
+
     let coords = sorted_coords::<FD, FS, D, T, C>(&txn, source.shape().clone(), coords).await?;
 
     let buffered = coords
@@ -87,6 +90,7 @@ where
     S: Stream<Item = TCResult<Coords>> + Send + Unpin,
 {
     debug!("sort coords for Tensor with shape {}", shape);
+
     let blocks = coords_to_offsets(shape, coords).map_ok(|block| ArrayExt::from(block).into());
 
     let block_list =
@@ -101,7 +105,12 @@ fn coords_to_offsets<S: Stream<Item = TCResult<Coords>> + Unpin>(
     shape: Shape,
     coords: S,
 ) -> impl Stream<Item = TCResult<Offsets>> {
-    coords.map_ok(move |coords| coords.to_offsets(&shape))
+    debug!("coords to offsets with shape {}", shape);
+
+    coords.map_ok(move |coords| {
+        debug_assert_eq!(coords.ndim(), shape.len());
+        coords.to_offsets(&shape)
+    })
 }
 
 fn offsets_to_coords<'a, S: Stream<Item = TCResult<Offsets>> + Unpin + 'a>(
