@@ -2,6 +2,7 @@ use std::collections::HashMap;
 use std::iter;
 
 use afarray::Coords;
+use log::debug;
 
 use tc_error::*;
 use tcgeneric::Tuple;
@@ -21,14 +22,22 @@ pub struct Broadcast {
 
 impl Broadcast {
     pub fn new(source_shape: Shape, shape: Shape) -> TCResult<Broadcast> {
-        let ndim = shape.len();
-        if source_shape.len() > shape.len() {
-            return Err(TCError::bad_request(
-                &format!("cannot broadcast into {}", shape),
-                source_shape,
+        if source_shape.is_empty() {
+            return Err(TCError::unsupported("cannot broadcast an empty Tensor"));
+        } else if shape.is_empty() {
+            return Err(TCError::unsupported(
+                "cannot broadcast into an empty Tensor",
             ));
         }
 
+        let ndim = shape.len();
+        if source_shape.len() > ndim {
+            return Err(TCError::unsupported(format!(
+                "cannot broadcast {} into {}",
+                source_shape, shape
+            )));
+        }
+
         let offset = ndim - source_shape.len();
         let mut inverted_axes = Vec::with_capacity(shape.len());
         let mut broadcast: Vec<bool> = iter::repeat(true).take(ndim).collect();
@@ -59,7 +68,7 @@ impl Broadcast {
     pub fn map_bounds(&self, source_bounds: Bounds) -> Bounds {
         assert_eq!(source_bounds.len(), self.source_shape.len());
 
-        let mut bounds = Bounds::all(&self.shape);
+        let mut bounds = Bounds::all(self.shape());
 
         for axis in 0..self.source_shape.len() {
             if !self.broadcast[axis + self.offset] {
@@ -78,10 +87,18 @@ impl Broadcast {
         let source_ndim = self.source_shape.len();
         let mut source_bounds = Vec::with_capacity(source_ndim);
         for axis in 0..source_ndim {
-            if self.broadcast[axis + self.offset] {
-                source_bounds.push(AxisBounds::from(0))
+            if axis + self.offset < bounds.len() {
+                if self.broadcast[axis + self.offset] {
+                    if bounds[axis + self.offset].is_index() {
+                        source_bounds.push(AxisBounds::from(0))
+                    } else {
+                        source_bounds.push(AxisBounds::In(0..1))
+                    }
+                } else {
+                    source_bounds.push(bounds[axis + self.offset].clone())
+                }
             } else {
-                source_bounds.push(bounds[axis + self.offset].clone())
+                source_bounds.push(AxisBounds::In(0..self.source_shape[axis]))
             }
         }
 
@@ -89,6 +106,8 @@ impl Broadcast {
     }
 
     pub fn invert_coord(&self, coord: &[u64]) -> Coord {
+        debug_assert_eq!(coord.len(), self.shape.len());
+
         let source_ndim = self.source_shape.len();
         let mut source_coord = Vec::with_capacity(source_ndim);
         for axis in 0..source_ndim {
@@ -117,7 +136,6 @@ pub struct Expand {
     source_shape: Shape,
     shape: Shape,
     expand: usize,
-    inverted_axes: Vec<usize>,
 }
 
 impl Expand {
@@ -126,28 +144,33 @@ impl Expand {
             return Err(TCError::bad_request("axis out of bounds", expand));
         }
 
-        let mut inverted_axes = Vec::with_capacity(source_shape.len() + 1);
-        inverted_axes.extend(0..source_shape.len());
-        inverted_axes.insert(expand, expand);
-
         let mut shape = source_shape.to_vec();
         shape.insert(expand, 1);
-        let shape: Shape = shape.into();
 
         Ok(Expand {
             source_shape,
-            shape,
+            shape: shape.into(),
             expand,
-            inverted_axes,
         })
     }
 
+    #[inline]
     pub fn expand_axis(&self) -> usize {
         self.expand
     }
 
     pub fn invert_axes(&self, axes: Vec<usize>) -> Vec<usize> {
-        axes.into_iter().map(|x| self.inverted_axes[x]).collect()
+        axes.into_iter()
+            .filter_map(|x| {
+                if x == self.expand {
+                    None
+                } else if x > self.expand {
+                    Some(x - 1)
+                } else {
+                    Some(x)
+                }
+            })
+            .collect()
     }
 
     pub fn shape(&'_ self) -> &'_ Shape {
@@ -155,15 +178,34 @@ impl Expand {
     }
 
     pub fn invert_bounds(&self, mut bounds: Bounds) -> Bounds {
-        if bounds.len() < self.expand {
-            bounds.remove(self.expand);
+        assert_eq!(bounds.len(), self.shape.len());
+
+        if self.expand < bounds.len() {
+            let removed = bounds.remove(self.expand);
+            if !removed.is_index() {
+                if self.expand == bounds.len() {
+                    let bound = match bounds.pop().unwrap() {
+                        AxisBounds::At(i) => AxisBounds::In(i..i + 1),
+                        other => other,
+                    };
+
+                    bounds.push(bound);
+                } else {
+                    let bound = match bounds.remove(self.expand) {
+                        AxisBounds::At(i) => AxisBounds::In(i..i + 1),
+                        other => other,
+                    };
+
+                    bounds.insert(self.expand, bound);
+                }
+            }
         }
 
         bounds
     }
 
     pub fn invert_coord(&self, coord: &[u64]) -> Coord {
-        assert_eq!(coord.len(), self.shape.len());
+        debug_assert_eq!(coord.len(), self.shape.len());
 
         let mut inverted = Vec::with_capacity(self.source_shape.len());
         inverted.extend(&coord[..self.expand]);
@@ -211,10 +253,6 @@ impl Reduce {
         })
     }
 
-    pub fn axis(&self) -> usize {
-        self.axis
-    }
-
     pub fn shape(&'_ self) -> &'_ Shape {
         &self.shape
     }
@@ -266,7 +304,7 @@ impl Slice {
     pub fn new(source_shape: Shape, bounds: Bounds) -> TCResult<Slice> {
         source_shape.validate_bounds(&bounds)?;
 
-        let mut shape: Coord = Vec::with_capacity(bounds.len());
+        let mut shape: Coord = Vec::with_capacity(source_shape.len());
         let mut offset = HashMap::new();
         let mut elided = HashMap::new();
         let mut inverted_axes = Vec::with_capacity(bounds.len());
@@ -331,34 +369,42 @@ impl Slice {
 
         let mut source_bounds = Vec::with_capacity(self.source_shape.len());
         let mut source_axis = 0;
-        for axis in 0..self.shape.len() {
-            if let Some(c) = self.elided.get(&axis) {
+        let mut axis = 0;
+        while source_axis < self.source_shape.len() {
+            if let Some(c) = self.elided.get(&source_axis) {
+                source_axis += 1;
                 source_bounds.push(AxisBounds::At(*c));
                 continue;
             }
 
             use AxisBounds::*;
-            match &bounds[source_axis] {
+            match &bounds[axis] {
                 In(range) => {
-                    if let In(source_range) = &self.bounds[axis] {
-                        let start = range.start + source_range.start;
-                        let end = start + (range.end - range.start);
-                        source_bounds.push((start..end).into());
+                    if source_axis < self.bounds.len() {
+                        if let In(source_range) = &self.bounds[source_axis] {
+                            let start = range.start + source_range.start;
+                            let end = start + (range.end - range.start);
+                            source_bounds.push((start..end).into());
+                        } else {
+                            assert_eq!(range.start, 0);
+                            source_bounds.push(self.bounds[source_axis].clone());
+                        }
                     } else {
-                        assert_eq!(range.start, 0);
-                        source_bounds.push(self.bounds[axis].clone());
+                        source_bounds.push(In(range.clone()));
                     }
                 }
                 Of(indices) => {
-                    let offset = self.offset.get(&axis).unwrap_or(&0);
+                    let offset = self.offset.get(&source_axis).unwrap_or(&0);
                     source_bounds.push(indices.iter().map(|i| i + offset).collect::<Coord>().into())
                 }
                 At(i) => {
-                    let offset = self.offset.get(&axis).unwrap_or(&0);
+                    let offset = self.offset.get(&source_axis).unwrap_or(&0);
                     source_bounds.push((i + offset).into())
                 }
             }
+
             source_axis += 1;
+            axis += 1;
         }
 
         source_bounds.into()
@@ -402,6 +448,11 @@ impl Slice {
 
         coord
     }
+
+    pub fn map_coords(&self, source_coords: Coords) -> Coords {
+        assert_eq!(source_coords.ndim(), self.source_shape.len());
+        source_coords.slice(&self.shape, &self.elided, &self.offset)
+    }
 }
 
 #[derive(Clone)]
@@ -415,11 +466,7 @@ pub struct Transpose {
 impl Transpose {
     pub fn new(source_shape: Shape, permutation: Option<Vec<usize>>) -> TCResult<Transpose> {
         let ndim = source_shape.len();
-        let permutation = permutation.unwrap_or_else(|| {
-            let mut axes: Vec<usize> = (0..ndim).collect();
-            axes.reverse();
-            axes
-        });
+        let permutation = permutation.unwrap_or_else(|| (0..ndim).rev().collect());
 
         if permutation.len() != ndim {
             return Err(TCError::unsupported(format!(
@@ -427,6 +474,13 @@ impl Transpose {
                 source_shape,
                 Tuple::from(permutation)
             )));
+        } else if let Some(max_axis) = permutation.iter().max() {
+            if max_axis >= &ndim {
+                return Err(TCError::bad_request(
+                    "cannot transpose nonexistent axis",
+                    max_axis,
+                ));
+            }
         }
 
         let mut shape: Coord = Vec::with_capacity(ndim);
@@ -462,7 +516,7 @@ impl Transpose {
         &self.shape
     }
 
-    pub fn invert_bounds(&self, bounds: Bounds) -> Bounds {
+    pub fn invert_bounds(&self, bounds: &Bounds) -> Bounds {
         let mut source_bounds = Bounds::all(&self.source_shape);
         for axis in 0..bounds.len() {
             source_bounds[self.permutation[axis]] = bounds[axis].clone();
@@ -487,26 +541,35 @@ impl Transpose {
     }
 
     pub fn invert_permutation(&self, bounds: &Bounds) -> Vec<usize> {
-        let mut permutation = Vec::with_capacity(self.shape.len());
+        debug!(
+            "source permutation is {:?}, bounds are {}",
+            self.permutation, bounds
+        );
+
+        let mut offset = 0;
+        let mut offsets = Vec::with_capacity(self.shape.len());
+        for bound in self.invert_bounds(bounds).iter() {
+            if bound.is_index() {
+                offset += 1;
+            }
+
+            offsets.push(offset);
+        }
+
+        let mut permutation = Vec::with_capacity(self.permutation.len());
         for (i, bound) in bounds.iter().enumerate() {
             if bound.is_index() {
                 // pass
             } else {
-                permutation.push(self.permutation[i]);
+                permutation.push(self.permutation[i] - offsets[self.permutation[i]]);
             }
         }
-        permutation
-    }
 
-    pub fn map_coord(&self, source_coord: Coord) -> Coord {
-        assert_eq!(source_coord.len(), self.permutation.len());
-
-        let mut coord = vec![0; source_coord.len()];
-        for axis in 0..source_coord.len() {
-            coord[self.permutation[axis]] = source_coord[axis];
+        for i in bounds.len()..self.shape.len() {
+            permutation.push(self.permutation[i] - offsets[self.permutation[i]]);
         }
 
-        coord
+        permutation
     }
 
     pub fn map_coords(&self, coords: Coords) -> Coords {
@@ -514,3 +577,93 @@ impl Transpose {
         coords.transpose(Some(&self.permutation))
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_broadcast_invert_bounds() {
+        let shape = Shape::from(vec![2, 3, 4, 1]);
+        let rebase = Broadcast::new(shape.clone(), vec![5, 6, 2, 3, 4, 10].into()).unwrap();
+        assert_eq!(
+            rebase.invert_bounds(vec![AxisBounds::In(0..1)].into()),
+            Bounds::all(&shape)
+        )
+    }
+
+    #[test]
+    fn test_slice_invert_bounds() {
+        let rebase = Slice::new(vec![2, 3, 4, 5].into(), Bounds::from(vec![0])).unwrap();
+        assert_eq!(rebase.shape().to_vec(), vec![3, 4, 5]);
+        assert_eq!(
+            rebase.invert_bounds(Bounds::from(vec![
+                AxisBounds::In(0..3),
+                AxisBounds::In(0..4),
+                AxisBounds::At(1)
+            ])),
+            Bounds::from(vec![
+                AxisBounds::At(0),
+                AxisBounds::In(0..3),
+                AxisBounds::In(0..4),
+                AxisBounds::At(1)
+            ])
+        );
+
+        let rebase = Slice::new(
+            vec![2, 3, 4, 5].into(),
+            Bounds::from(vec![
+                AxisBounds::At(0),
+                AxisBounds::In(0..3),
+                AxisBounds::In(1..3),
+                AxisBounds::At(1),
+            ]),
+        )
+        .unwrap();
+
+        assert_eq!(
+            rebase.invert_bounds(Bounds::from(vec![AxisBounds::At(0), AxisBounds::In(0..2)])),
+            Bounds::from(vec![
+                AxisBounds::At(0),
+                AxisBounds::At(0),
+                AxisBounds::In(1..3),
+                AxisBounds::At(1)
+            ])
+        );
+    }
+
+    #[test]
+    fn test_transpose_invert_permutation() {
+        let rebase = Transpose::new(vec![10, 15, 20].into(), Some(vec![0, 1, 2])).unwrap();
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::At(0), AxisBounds::In(2..5)])),
+            vec![0, 1]
+        );
+
+        let rebase = Transpose::new(vec![10, 15, 20].into(), None).unwrap();
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::At(0), AxisBounds::In(2..5)])),
+            vec![1, 0]
+        );
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::In(0..2), AxisBounds::At(1)])),
+            vec![1, 0]
+        );
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::In(0..2)])),
+            vec![2, 1, 0]
+        );
+
+        let rebase = Transpose::new(vec![10, 15, 20, 25].into(), None).unwrap();
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::In(0..2), AxisBounds::At(1)])),
+            vec![2, 1, 0]
+        );
+
+        let rebase = Transpose::new(vec![10, 15, 20, 25].into(), Some(vec![3, 0, 1, 2])).unwrap();
+        assert_eq!(
+            rebase.invert_permutation(&Bounds::from(vec![AxisBounds::In(0..2), AxisBounds::At(1)])),
+            vec![2, 0, 1]
+        );
+    }
+}
